# -*- coding=utf-8 -*-
r"""

"""
import shlex
import typing as t
from .typing import KEY, INDEX


def unify_key(key: KEY) -> t.Tuple[INDEX, ...]:
    if isinstance(key, slice):
        return tuple(k for k in (key.start, key.stop, key.step) if k is not None)
    elif isinstance(key, (tuple, list)):
        return tuple(key)
    else:
        return (key,)


class Convert:
    BOOLEAN_STATES = {'1': True, 'yes': True, 'true': True, 'on': True,
                      '0': False, 'no': False, 'false': False, 'off': False}

    @staticmethod
    def to_str(value: t.Any):
        return str(value)

    @staticmethod
    def to_int(value: t.Any):
        return int(value)

    @staticmethod
    def to_float(value: t.Any):
        return float(value)

    @staticmethod
    def to_bool(value: t.Any):
        if isinstance(value, bool):
            return value
        elif isinstance(value, str):
            return Convert.BOOLEAN_STATES[value]
        elif value in {0, 1}:
            return bool(value)
        else:
            raise ValueError(f"can't safely ensure bool for {value!r} of type {type(value).__name__}")

    @staticmethod
    def split(value: t.Any):
        if isinstance(value, (tuple, list)):
            return list(map(str, value))
        elif isinstance(value, str):
            import re
            return re.split(r'\s*[,;]\s*', value)
        else:
            raise TypeError(f"can't split value of type {type(value).__name__}")

    @staticmethod
    def split_paths(value: t.Any):
        if isinstance(value, (tuple, list)):
            return list(map(str, value))
        elif isinstance(value, str):
            import os.path as p
            return value.split(p.pathsep)
        else:
            raise TypeError(f"can't path-split value of type {type(value).__name__}")

    @staticmethod
    def split_shlex(value: t.Any):
        return shlex.split(str(value))



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
import typing as t
import warnings


__all__ = ["register", "unregister"]


_T = t.TypeVar("_T")
_P = t.ParamSpec("_P")
__functions: t.Dict[t.Callable[_P, _T], t.Tuple[_P.args, _P.kwargs]] = dict()


def register(fn: t.Callable[_P, _T], *args: _P.args, **kwargs: _P.kwargs):
    __functions[fn] = (args, kwargs)
    return fn


def unregister(fn: t.Callable[_P, _T]):
    del __functions[fn]


def _run_restartfuncs():
    for fn, (args, kwargs) in __functions.items():
        try:
            fn(*args, **kwargs)
        except BaseException as exc:
            warnings.warn(f"atrestart functions failed with {type(exc).__name__} ({exc})", RuntimeWarning)



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
from pathlib import Path
from ..exceptions import ConfigNotFoundError
from ..loader import get_supported_formats
from .default_places import DEFAULT_PLACES, T_PLACES


def find(*variants: str, places: T_PLACES = None) -> Path:
    r"""
    find a config file

    note: if a variant ends with .ext then every supported config-type is searched for

    :param variants: name-variants of the config file
    :param places: list of directories to search in
    :return: pathlib.Path
    :raise ConfigNotFoundError: config-file could not be found
    """
    if places is None:
        places = DEFAULT_PLACES

    # Todo. Split variants by absolute or not

    for place in places:
        if callable(place):
            place = place()
        if place is None:
            continue

        place = Path(place)
        if not place.is_dir():
            continue

        for variant in variants:
            fp = place / variant
            if fp.suffix == ".ext":
                for extension in get_supported_formats():
                    fpx = fp.with_suffix(f".{extension}")
                    if fpx.is_file():
                        return fpx
            else:
                if fp.is_file():
                    return fp

    # Error Message
    raise ConfigNotFoundError("no config-file could be found anywhere")



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
import os
import typing as t
from ..exceptions import NotSupportedError
from ..interface import ConfigInterface
from .registry import REGISTRY, register_loader
from .environ import load_env
from . import loaders


__all__ = ["NotSupportedError", "REGISTRY", "register_loader", "get_supported_formats", "load"]


def get_supported_formats() -> t.Iterable[str]:
    r"""
    return currently supported file extensions

    eg: ini, conf, json
    """
    return tuple(REGISTRY.keys())


def load(fp: t.Union[str, os.PathLike]) -> ConfigInterface:
    r"""
    find the correct loader for the passed file, parses it and returns the result

    :param fp: path of the config-file
    :return: depends on the file-extension
    """
    import os.path as p
    ext = p.splitext(fp)[1][1:]
    if ext not in REGISTRY:
        raise NotSupportedError(f"files of type .{ext} are not supported")
    loader = REGISTRY[ext]
    return ConfigInterface(loader(fp))



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""
MIT License

Copyright (c) 2023 PlayerG9

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

__author__ = "PlayerG9"
__copyright__ = "Copyright 2023, PlayerG9"
__credits__ = ["PlayerG9"]
__license__ = "MIT"
__maintainer__ = "PlayerG9"
__email__ = None
__status__ = "Prototype"  # Prototype, Development, Production
__description__ = "utility library to find and load configuration files"
from .__version__ import __version__, __version_info__

from .exceptions import *
from .finder import find
from .loader import load, get_supported_formats, loaders
from .interface import ConfigInterface


def find_and_load(*variants: str, places=None) -> ConfigInterface:
    r"""
    common interface for the `find()` and `load()` function

    basically an alias for `load(find(...))`

    :param variants: same as `find()`: name-variants of the config file
    :param places: same as `find()`: list of directories to search in
    :return: ConfigInterface
    """
    fp = find(*variants, places=places)
    return load(fp)


def load_environ(prefix: str) -> ConfigInterface:
    r"""
    loads the environment variables into an object.
    The object gets nested by splitting the environment keys on `__`

    :param prefix: app prefix to filter the environment variables
    :return: ConfigInterface
    """
    from .loader.environ import load_env
    return ConfigInterface(load_env(prefix=prefix))



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
import os
import typing as t
from configlib.exceptions import NotSupportedError
from configlib.loader import register_loader
try:
    # noinspection PyCompatibility
    import tomllib  # python 3.11+
except ModuleNotFoundError:
    try:
        import toml as tomllib  # pip install config-library[toml]
    except ModuleNotFoundError:
        raise NotSupportedError('please install config-library[toml] for this')


@register_loader('toml')
def load_toml(fp: t.Union[str, os.PathLike]) -> dict:
    r"""
    # This is a TOML document

    title = "TOML Example"

    [owner]
    name = "Tom Preston-Werner"
    dob = 1979-05-27T07:32:00-08:00

    [database]
    enabled = true
    ports = [ 8000, 8001, 8002 ]
    data = [ ["delta", "phi"], [3.14] ]
    temp_targets = { cpu = 79.5, case = 72.0 }

    [servers]

    [servers.alpha]
    ip = "10.0.0.1"
    role = "frontend"

    [servers.beta]
    ip = "10.0.0.2"
    role = "backend"
    """
    with open(fp, 'r') as file:
        return tomllib.load(file)



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""
https://github.com/muhammadmuzzammil1998/jsonc
"""
import os
import re
import json
import typing as t
from configlib.loader import register_loader


_REGEX = re.compile(r"(\".*?\"|\'.*?\')|(/\*.*?\*/|//[^\r\n]*$)", re.MULTILINE | re.DOTALL)


@register_loader('jsonc')
def load_jsonc(fp: t.Union[str, os.PathLike]) -> dict:
    r"""
    {
        /* This is an example
           for block comment.  */
        "foot": "bar foo",  // Comments
        "true": false,      // Improve readability
        "number": 42,       // Number will always be 42
        /* Comments are ignored while
           generating JSON from JSONC. */
        // "object": {
        //   "test": "done"
        // }
        "array": [1, 2, 3]
    }
    """

    def __replace(match):
        return "" if match.group(2) is not None else match.group(1)

    with open(fp, 'r') as file:
        content = file.read()

    content = _REGEX.sub(__replace, content)

    return json.loads(content)



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
import sys
import os.path as p
from ..interface import ConfigInterface


ModuleClass = type(sys)


def install_config(config, *, fp: str = None):
    r"""

    :param config: the config-object to translate
    :param fp: set __file__ attribute
    :return:
    """

    if 'config' in sys.modules:
        raise ValueError("'config' module already exists")

    # note: sets __name__ and __doc__
    module = ModuleClass(name="config", doc="automatically generated module bases on one object")

    if fp:
        module.__file__ = p.abspath(fp)

    if isinstance(config, ConfigInterface):
        config = config.get()
    if isinstance(config, dict):
        for key in config:
            setattr(module, key, config[key])
    else:
        for key in dir(config):
            setattr(module, key, getattr(config, key))

    sys.modules['config'] = module


def reinstall_config(config, *, fp: str = None):
    r"""

    :param config: the config-object to translate
    :param fp: set __file__ attribute
    :return:
    """
    del sys.modules['config']
    install_config(config=config, fp=fp)



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
import os
import typing as t
from configlib.exceptions import NotSupportedError
from configlib.loader import register_loader
try:
    import yaml
except ModuleNotFoundError:
    raise NotSupportedError('please install config-library[yaml] for this')


@register_loader('yaml', 'yml')
def load_yaml(fp: t.Union[str, os.PathLike]) -> dict:
    r"""
    receipt:     Oz-Ware Purchase Invoice
    date:        2012-08-06
    customer:
        first_name:   Dorothy
        family_name:  Gale

    items:
        - part_no:   A4786
          descrip:   Water Bucket (Filled)
          price:     1.47
          quantity:  4

        - part_no:   E1628
          descrip:   High Heeled "Ruby" Slippers
          size:      8
          price:     133.7
          quantity:  1

    bill-to:  &id001
        street: |
                123 Tornado Alley
                Suite 16
        city:   East Centerville
        state:  KS

    ship-to:  *id001

    specialDelivery:  >
        Follow the Yellow Brick
        Road to the Emerald City.
        Pay no attention to the
        man behind the curtain.
    """
    with open(fp, 'r') as file:
        return yaml.safe_load(file)



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
import os
import typing as t
import warnings


T_LOADER: t.TypeAlias = t.Callable[[t.Union[str, os.PathLike]], t.Dict]
T = t.TypeVar('T')


REGISTRY: t.Dict[str, T_LOADER] = {}


def register_loader(*extensions: str) -> t.Callable[[T], T]:
    r"""
    register function as loader for 1+ extensions

    @register_loader('ext')
    def load_exc(fp: str) -> Any:
        ...
    """
    if not extensions:
        raise ValueError("Please provide at least on extension")

    def add_to_registry(fn: T) -> T:
        for extension in extensions:
            ext = extension.removeprefix('.')
            if ext in REGISTRY:
                warnings.warn(f"replace loader for .{ext} files")
            REGISTRY[ext] = fn
        return fn

    return add_to_registry



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
import sys; sys.path.append('./src')  # noqa
import setuptools
from configlib import __author__, __version__, __description__, __license__


install_requires = []

yaml_requires = ["pyyaml"]
toml_requires = ["toml"]
json5_requires = ["json-five"]
watcher_requires = ["watchdog"]
all_requires = [yaml_requires, toml_requires, json5_requires, watcher_requires]

extras_require = {
    'yaml': yaml_requires,
    'toml': toml_requires,
    'json5': json5_requires,
    'watcher': watcher_requires,
    'all': all_requires,
}

setuptools.setup(
    name="config-library",
    version=__version__,
    description=__description__,
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author=__author__,
    license=__license__,
    url="https://github.com/PlayerG9/configlib-py",
    project_urls={
        "Author Github": "https://github.com/PlayerG9",
        "Homepage": "https://github.com/PlayerG9/configlib-py/",
        "Documentation": "https://utility-libraries.github.io/configlib-py/",
        "Bug Tracker": "https://github.com/PlayerG9/configlib-py/issues",
    },
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Utilities",
    ],
    python_requires=">=3.5",
    install_requires=install_requires,
    extras_require=extras_require,
    test_suite="tests",
)



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
import os
import typing as t
from configlib.exceptions import NotSupportedError
from configlib.loader import register_loader
try:
    import json5
except ModuleNotFoundError:
    raise NotSupportedError('please install config-library[json5] for this')


@register_loader('json5')
def load_json5(fp: t.Union[str, os.PathLike]) -> dict:
    r"""
    {
      // comments
      unquoted: 'and you can quote me on that',
      singleQuotes: 'I can use "double quotes" here',
      lineBreaks: "Look, Mom! \
    No \\n's!",
      hexadecimal: 0xdecaf,
      leadingDecimalPoint: .8675309, andTrailing: 8675309.,
      positiveSign: +1,
      trailingComma: 'in objects', andIn: ['arrays',],
      "backwardsCompatible": "with JSON",
    }
    """
    with open(fp, 'r') as file:
        return json5.load(file)



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
import typing as t
from ..exceptions import CallOrderError


class Configurator:
    r"""
    utility combiner for configlib

        from configlib import Configurator
        Configurator('app.conf').find().load().install().make_restart_on_change()
        ...
        from config import HOST
    """

    def __init__(self, *variants: str, restart_on_change: bool = False):
        r"""
        :param name: name of the configuration file
        :param namespace: namespace of the project
        :param ns_only: only search for {namespace}/{name}
        :param restart_on_change: restart the whole application if the configuration file changes
        """
        self._variants = variants
        self._fp = None
        self._config = None
        self._watcher = None
        self.restart_on_change = restart_on_change
        self.run_atexit = False

    @property
    def variants(self) -> t.Tuple[str, ...]:
        return self._variants

    @property
    def file(self) -> t.Optional[str]:
        return self._fp

    @property
    def config(self) -> t.Any:
        if self._config is None:
            raise AttributeError(f"load() wasn't called yet")
        return self._config

    def find_and_load(self) -> 'Configurator':
        r"""
        find and load the configuration file
        """
        self.find()
        self.load()
        return self

    def find(self) -> 'Configurator':
        r"""
        find the configuration file
        """
        from ..finder import find
        self._fp = find(*self.variants)
        return self

    def load(self) -> 'Configurator':
        r"""
        load the configuration file
        """
        from ..loader import load
        if self.file is None:
            raise CallOrderError(f"load() has to be called after find()")
        self._config = load(self.file)
        return self

    def install(self) -> 'Configurator':
        r"""
        installed the loaded config file to be imported under 'config'

        see configlib.install
        """
        from ..install import install_config
        if self._config is None:
            raise CallOrderError(f"install() has to be called after load()")
        install_config(self._config)
        return self

    def make_restart_on_change(self, *, run_atexit: bool = None) -> 'Configurator':
        r"""
        restart the program on config-file changes
        """
        self.restart_on_change = True
        if run_atexit is not None:
            self.run_atexit = run_atexit
        self.watch()
        return self

    def _handle_change(self):
        from .restarter import restart
        if self.restart_on_change:
            restart(run_atexit=self.run_atexit)

    def watch(self) -> 'Configurator':
        r"""
        watch for changes to the configuration file and call handlers registered with .on_change()
        """
        from ..watcher import Watcher
        if self._watcher:
            return self
        if self._config is None:
            raise CallOrderError(f"watch() has to be called after find()")
        self._watcher = Watcher(self.file)
        self._watcher.start()
        self._watcher.on_change(self._handle_change)
        return self

    def stop_watching(self) -> 'Configurator':
        r"""
        stop watching for changes
        """
        if not self._watcher:
            return self
        self._watcher.stop()
        self._watcher.join()
        self._watcher = None
        return self

    def on_change(self, handler):
        r"""
        register handler to be called when the configuration file changes
        """
        return self._watcher.on_change(handler)



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""
# main.py
    import configlib
    configlib.install(dict(
        HOST="localhost",
        PORT=8000,
    ), fp=config_file)
    import sub

# config.pyi
    HOST: str
    PORT: int

# sub.py
    import config
    from config import HOST, PORT
    ...
"""
from .installer import install_config, reinstall_config



#!/usr/bin/python3
# -*- coding=utf-8 -*-
r"""

"""
import os
import sys
import atexit
from .atrestart import _run_restartfuncs


def restart(run_atexit: bool = False):
    r"""
    This function restarts your python program.
    PID and arguments stay the same.

    Warnings:
        - Stop further execution of code behind this function
        - It does not properly close files or other logic. So be careful.

    :param run_atexit: whether function registered with @atexit should be run or not
    """
    _run_restartfuncs()
    if run_atexit:
        atexit._run_exitfuncs()  # noqa
    os.execv(sys.executable, sys.orig_argv)



# -*- coding=utf-8 -*-
r"""

"""
import os
import typing as t
from .util import default_split, default_transform


def load_env(
        prefix: str,
        override: bool = True,
        transform: t.Optional[t.Callable[[str], str]] = default_transform,
        split: t.Optional[t.Callable[[str], t.List[str]]] = default_split,
) -> t.Dict[str, t.Union[str, dict]]:
    r"""
    information for override
    if 'HELLO' and 'HELLO__WORLD' exist, then 'HELLO' gets overriden and is not included

    :param prefix: prefix to filter for (_ is automatically added to the prefix)
    :param override: in case of a conflict in a nested object the existing gets overriden, otherwise KeyError
    :param transform: function to transform the key (default: HELLO-WORLD -> hello_world)
    :param split: functions used to split the key to create a nested object (default: split by __)
    :return: dict
    """
    if transform is None:
        def transform(n):
            return n
    if split is None:
        def split(s):
            return [s]

    prefix = f"{prefix}_"

    env_object = {}

    for raw_key in os.environ.keys():
        if not raw_key.startswith(prefix):
            continue
        key = transform(raw_key.removeprefix(prefix))
        *path, setter = split(key)
        ref = env_object
        for k in path:
            k = transform(k)
            if k not in ref:
                ref[k] = {}
            elif not isinstance(ref[k], dict):
                if override:
                    ref[k] = {}
                else:
                    raise KeyError(f"{k} of {'->'.join(path)} exist")
            ref = ref[k]
        ref[setter] = os.environ[raw_key]

    return env_object
